"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const grammarkdown_1 = require("grammarkdown");
const utils_1 = require("./utils");
function collectGrammarDiagnostics(dom, sourceText, mainGrammar, sdos, earlyErrors) {
    // *******************
    // Parse the grammar with Grammarkdown and collect its diagnostics, if any
    var _a;
    let grammarParser = new grammarkdown_1.Parser();
    // TODO use CoreSyncHost once published
    let fakeHost = {
        readFileSync(file) {
            let idx = parseInt(file);
            if (idx.toString() !== file || idx < 0 || idx >= mainGrammar.length) {
                throw new Error('tried to read non-existent ' + file);
            }
            return mainGrammar[idx].source;
        },
        resolveFile(file) {
            return file;
        },
        normalizeFile(file) {
            return file;
        },
        getSourceFileSync(file) {
            return grammarParser.parseSourceFile(file, this.readFileSync(file));
        },
    }; // good enough!
    let compilerOptions = {
        format: grammarkdown_1.EmitFormat.ecmarkup,
        noChecks: false,
        noUnusedParameters: true,
    };
    let grammar = new grammarkdown_1.Grammar(Object.keys(mainGrammar), compilerOptions, fakeHost);
    grammar.parseSync();
    grammar.checkSync();
    let lintingErrors = [];
    let unusedParameterErrors = new Map();
    if (grammar.diagnostics.size > 0) {
        // `detailedMessage: false` prevents prepending line numbers, which is good because we're going to make our own
        grammar.diagnostics
            .getDiagnosticInfos({ formatMessage: true, detailedMessage: false })
            .forEach(m => {
            let idx = +m.sourceFile.filename;
            let grammarLoc = utils_1.getLocation(dom, mainGrammar[idx].element);
            let { line, column } = utils_1.grammarkdownLocationToTrueLocation(grammarLoc, m.range.start.line, m.range.start.character);
            let error = {
                ruleId: `grammarkdown:${m.code}`,
                nodeType: m.node === undefined ? 'unknown' : grammarkdown_1.tokenToString(m.node.kind, false),
                line,
                column,
                message: m.formattedMessage,
            };
            lintingErrors.push(error);
            if (m.code === grammarkdown_1.Diagnostics.Parameter_0_is_unused.code) {
                let param = m.node;
                let navigator = grammar.resolver.createNavigator(param);
                navigator.moveToAncestor(node => node.kind === grammarkdown_1.SyntaxKind.Production);
                let nodeName = navigator.getNode().name.text;
                let paramName = param.name.text;
                if (!unusedParameterErrors.has(nodeName)) {
                    unusedParameterErrors.set(nodeName, new Map());
                }
                let paramToError = unusedParameterErrors.get(nodeName);
                paramToError.set(paramName, error);
            }
        });
    }
    // *******************
    // Check that SDOs and Early Errors are defined in terms of productions which actually exist
    // Also filter out any "unused parameter" warnings for grammar productions for which the parameter is used in an early error or SDO
    let oneOffGrammars = [];
    let actualGrammarProductions = utils_1.getProductions(grammar);
    let grammarsAndRules = [
        ...sdos.map(s => ({ grammar: s.grammar, rules: [s.alg], type: 'syntax-directed operation' })),
        ...earlyErrors.map(e => ({ grammar: e.grammar, rules: e.lists, type: 'early error' })),
    ];
    for (let { grammar: grammarEle, rules: rulesEles, type } of grammarsAndRules) {
        const nodeType = grammarEle.tagName;
        let grammarLoc = utils_1.getLocation(dom, grammarEle);
        if (grammarLoc.endTag == null) {
            lintingErrors.push({
                ruleId: 'missing-close-tag',
                message: 'could not find closing tag for emu-grammar',
                line: grammarLoc.startTag.line,
                column: grammarLoc.startTag.col,
                nodeType: 'EMU-GRAMMAR',
            });
            continue;
        }
        let grammarHost = grammarkdown_1.SyncHost.forFile(sourceText.slice(grammarLoc.startTag.endOffset, grammarLoc.endTag.startOffset));
        let grammar = new grammarkdown_1.Grammar([grammarHost.file], {}, grammarHost);
        grammar.parseSync();
        oneOffGrammars.push({ grammarEle, grammar });
        let productions = utils_1.getProductions(grammar);
        function getLocationInGrammar(pos) {
            let file = grammar.sourceFiles[0];
            let posWithoutWhitespace = grammarkdown_1.skipTrivia(file.text, pos, file.text.length);
            let { line: gmdLine, character: gmdCharacter } = file.lineMap.positionAt(posWithoutWhitespace);
            return utils_1.grammarkdownLocationToTrueLocation(grammarLoc, gmdLine, gmdCharacter);
        }
        for (let [name, { production, rhses }] of productions) {
            let originalRhses = (_a = actualGrammarProductions.get(name)) === null || _a === void 0 ? void 0 : _a.rhses;
            if (originalRhses === undefined) {
                let { line, column } = getLocationInGrammar(production.pos);
                lintingErrors.push({
                    ruleId: 'undefined-nonterminal',
                    nodeType,
                    line,
                    column,
                    message: `Could not find a definition for LHS in ${type}`,
                });
                continue;
            }
            for (let rhs of rhses) {
                if (!originalRhses.some(o => utils_1.rhsMatches(rhs, o))) {
                    let { line, column } = getLocationInGrammar(rhs.pos);
                    lintingErrors.push({
                        ruleId: 'undefined-nonterminal',
                        nodeType,
                        line,
                        column,
                        message: `Could not find a production matching RHS in ${type}`,
                    });
                }
                if (rhs.kind === grammarkdown_1.SyntaxKind.RightHandSide) {
                    (function noGrammarRestrictions(s) {
                        if (s === undefined) {
                            return;
                        }
                        if (s.symbol.kind === grammarkdown_1.SyntaxKind.NoSymbolHereAssertion) {
                            let { line, column } = getLocationInGrammar(s.symbol.pos);
                            lintingErrors.push({
                                ruleId: `NLTH-in-SDO`,
                                nodeType,
                                line,
                                column,
                                message: `Productions referenced in ${type}s should not include "no LineTerminator here" restrictions`,
                            });
                        }
                        // We could also enforce that lookahead restrictions are absent, but in some cases they actually do add clarity, so we just don't enforce it either way.
                        noGrammarRestrictions(s.next);
                    })(rhs.head);
                    if (rhs.constraints !== undefined) {
                        let { line, column } = getLocationInGrammar(rhs.constraints.pos);
                        lintingErrors.push({
                            ruleId: `guard-in-SDO`,
                            nodeType,
                            line,
                            column,
                            message: `Productions referenced in ${type}s should not be gated on grammar parameters`,
                        });
                    }
                }
            }
            // Filter out unused parameter errors for which the parameter is actually used in an SDO or Early Error
            if (unusedParameterErrors.has(name)) {
                let paramToError = unusedParameterErrors.get(name);
                for (let [paramName, error] of paramToError) {
                    // This isn't the most elegant check, but it works.
                    if (rulesEles.some(r => r.innerHTML.indexOf('[' + paramName + ']') !== -1)) {
                        paramToError.delete(paramName);
                        // Yes, there's definitely big-O faster ways of doing this, but in practice this is probably faster for the sizes we will encounter.
                        let index = lintingErrors.indexOf(error);
                        if (index === -1) {
                            throw new Error('unreachable: tried to clear non-existent error');
                        }
                        lintingErrors.splice(index, 1);
                    }
                }
            }
        }
    }
    return {
        grammar,
        oneOffGrammars,
        lintingErrors,
    };
}
exports.collectGrammarDiagnostics = collectGrammarDiagnostics;
