"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("./utils");
function collectNodes(sourceText, dom, document) {
    let headers = [];
    let mainGrammar = [];
    let sdos = [];
    let earlyErrors = [];
    let algorithms = [];
    let failed = false;
    let errors = [];
    let inAnnexB = false;
    let lintWalker = document.createTreeWalker(document.body, 1 /* elements */);
    function visitCurrentNode() {
        var _a;
        let node = lintWalker.currentNode;
        let thisNodeIsAnnexB = node.nodeName === 'EMU-ANNEX' &&
            node.id === 'sec-additional-ecmascript-features-for-web-browsers';
        if (thisNodeIsAnnexB) {
            inAnnexB = true;
        }
        // Don't bother collecting early errors and SDOs from Annex B.
        // This is mostly so we don't have to deal with having two inconsistent copies of some of the grammar productions.
        if (!inAnnexB) {
            if (node.nodeName === 'EMU-CLAUSE') {
                // Look for early errors
                let first = node.firstElementChild;
                if (first !== null && first.nodeName === 'H1') {
                    let title = (_a = first.textContent) !== null && _a !== void 0 ? _a : '';
                    headers.push({ element: first, contents: title });
                    if (title.trim() === 'Static Semantics: Early Errors') {
                        let grammar = null;
                        let lists = [];
                        for (let child of node.children) {
                            if (child.nodeName === 'EMU-GRAMMAR') {
                                if (grammar !== null) {
                                    if (lists.length === 0) {
                                        throw new Error('unrecognized structure for early errors: grammar without errors');
                                    }
                                    earlyErrors.push({ grammar, lists });
                                }
                                grammar = child;
                                lists = [];
                            }
                            else if (child.nodeName === 'UL') {
                                if (grammar === null) {
                                    throw new Error('unrecognized structure for early errors: errors without correspondinig grammar');
                                }
                                lists.push(child);
                            }
                        }
                        if (grammar === null) {
                            throw new Error('unrecognized structure for early errors: no grammars');
                        }
                        if (lists.length === 0) {
                            throw new Error('unrecognized structure for early errors: grammar without errors');
                        }
                        earlyErrors.push({ grammar, lists });
                    }
                }
            }
            else if (node.nodeName === 'EMU-GRAMMAR') {
                // Look for grammar definitions and SDOs
                if (node.getAttribute('type') === 'definition') {
                    let loc = utils_1.getLocation(dom, node);
                    if (loc.endTag == null) {
                        failed = true;
                        errors.push({
                            ruleId: 'missing-close-tag',
                            message: 'could not find closing tag for emu-grammar',
                            line: loc.startTag.line,
                            column: loc.startTag.col,
                            nodeType: 'EMU-GRAMMAR',
                        });
                    }
                    else {
                        let start = loc.startTag.endOffset;
                        let end = loc.endTag.startOffset;
                        let realSource = sourceText.slice(start, end);
                        mainGrammar.push({ element: node, source: realSource });
                    }
                }
                else if (node.getAttribute('type') !== 'example') {
                    let next = lintWalker.nextSibling();
                    if (next) {
                        if (next.nodeName === 'EMU-ALG') {
                            sdos.push({ grammar: node, alg: next });
                        }
                        lintWalker.previousSibling();
                    }
                }
            }
        }
        if (node.nodeName === 'EMU-ALG' && node.getAttribute('type') !== 'example') {
            algorithms.push({ element: node });
        }
        let firstChild = lintWalker.firstChild();
        if (firstChild) {
            while (true) {
                visitCurrentNode();
                let next = lintWalker.nextSibling();
                if (!next)
                    break;
            }
            lintWalker.parentNode();
        }
        if (thisNodeIsAnnexB) {
            inAnnexB = false;
        }
    }
    visitCurrentNode();
    if (failed) {
        return { success: false, errors };
    }
    return { success: true, mainGrammar, headers, sdos, earlyErrors, algorithms };
}
exports.collectNodes = collectNodes;
